@@ -3,7 +3,8 @@
  */
 
 /* 
- * Copyright (C) 1986, 1988, 1989, 1991-2001, 2003-2009 the Free Software Foundation, Inc.
+ * Copyright (C) 1986, 1988, 1989, 1991-2001, 2003-2015,
+ * the Free Software Foundation, Inc.
  * 
  * This file is part of GAWK, the GNU implementation of the
  * AWK Programming Language.
@@ -25,32 +26,32 @@
 
 #include "awk.h"
 #include "math.h"
+#include "floatmagic.h"	/* definition of isnan */
 
-static int is_ieee_magic_val P((const char *val));
-static AWKNUM get_ieee_magic_val P((const char *val));
+static int is_ieee_magic_val(const char *val);
+static NODE *r_make_number(double x);
+static AWKNUM get_ieee_magic_val(const char *val);
+extern NODE **fmt_list;          /* declared in eval.c */
 
-/* r_force_number --- force a value to be numeric */
+NODE *(*make_number)(double) = r_make_number;
+NODE *(*str2number)(NODE *) = r_force_number;
+NODE *(*format_val)(const char *, int, NODE *) = r_format_val;
+int (*cmp_numbers)(const NODE *, const NODE *) = cmp_awknums;
 
-AWKNUM
-r_force_number(register NODE *n)
+/* force_number --- force a value to be numeric */
+
+NODE *
+r_force_number(NODE *n)
 {
-	register char *cp;
-	register char *cpend;
+	char *cp;
+	char *cpend;
 	char save;
 	char *ptr;
 	unsigned int newflags;
 	extern double strtod();
 
-#ifdef GAWKDEBUG
-	if (n == NULL)
-		cant_happen();
-	if (n->type != Node_val)
-		cant_happen();
-	if (n->flags == 0)
-		cant_happen();
-	if (n->flags & NUMCUR)
-		return n->numbr;
-#endif
+	if ((n->flags & NUMCUR) != 0)
+		return n;
 
 	/* all the conditionals are an attempt to avoid the expensive strtod */
 
@@ -59,9 +60,7 @@ r_force_number(register NODE *n)
 	n->numbr = 0.0;
 
 	if (n->stlen == 0) {
-		if (0 && do_lint)
-			lintwarn(_("can't convert string to float"));
-		return 0.0;
+		return n;
 	}
 
 	cp = n->stptr;
@@ -73,17 +72,15 @@ r_force_number(register NODE *n)
 	 * This also allows hexadecimal floating point. Ugh.
 	 */
 	if (! do_posix) {
-		if (ISALPHA(*cp)) {
-			if (0 && do_lint)
-				lintwarn(_("can't convert string to float"));
-			return 0.0;
+		if (is_alpha((unsigned char) *cp)) {
+			return n;
 		} else if (n->stlen == 4 && is_ieee_magic_val(n->stptr)) {
-			if (n->flags & MAYBE_NUM)
+			if ((n->flags & MAYBE_NUM) != 0)
 				n->flags &= ~MAYBE_NUM;
 			n->flags |= NUMBER|NUMCUR;
 			n->numbr = get_ieee_magic_val(n->stptr);
 
-			return n->numbr;
+			return n;
 		}
 		/* else
 			fall through */
@@ -92,39 +89,38 @@ r_force_number(register NODE *n)
 		fall through */
 
 	cpend = cp + n->stlen;
-	while (cp < cpend && ISSPACE(*cp))
+	while (cp < cpend && isspace((unsigned char) *cp))
 		cp++;
 
-	/* FIXME: Simplify this condition! */
-	if (   cp == cpend
-	    || (! do_posix
-	        && (ISALPHA(*cp)
+	if (   cp == cpend		/* only spaces, or */
+	    || (! do_posix		/* not POSIXLY paranoid and */
+	        && (is_alpha((unsigned char) *cp)	/* letter, or */
+					/* CANNOT do non-decimal and saw 0x */
 		    || (! do_non_decimal_data && cp[0] == '0'
 		        && (cp[1] == 'x' || cp[1] == 'X'))))) {
-		if (0 && do_lint)
-			lintwarn(_("can't convert string to float"));
-		return 0.0;
+		return n;
 	}
 
-	if (n->flags & MAYBE_NUM) {
+	if ((n->flags & MAYBE_NUM) != 0) {
 		newflags = NUMBER;
 		n->flags &= ~MAYBE_NUM;
 	} else
 		newflags = 0;
 
-	if (cpend - cp == 1) {
-		if (ISDIGIT(*cp)) {
+	if (cpend - cp == 1) {		/* only one character */
+		if (isdigit((unsigned char) *cp)) {	/* it's a digit! */
 			n->numbr = (AWKNUM)(*cp - '0');
 			n->flags |= newflags;
 			n->flags |= NUMCUR;
-		} else if (0 && do_lint)
-			lintwarn(_("can't convert string to float"));
-		return n->numbr;
+			if (cp == n->stptr)		/* no leading spaces */
+				n->flags |= NUMINT;
+		}
+		return n;
 	}
 
 	if (do_non_decimal_data) {	/* main.c assures false if do_posix */
 		errno = 0;
-		if (! do_traditional && isnondecimal(cp, TRUE)) {
+		if (! do_traditional && get_numbase(cp, true) != 10) {
 			n->numbr = nondec2awknum(cp, cpend - cp);
 			n->flags |= NUMCUR;
 			ptr = cpend;
@@ -138,28 +134,34 @@ r_force_number(register NODE *n)
 	n->numbr = (AWKNUM) strtod((const char *) cp, &ptr);
 
 	/* POSIX says trailing space is OK for NUMBER */
-	while (ISSPACE(*ptr))
+	while (isspace((unsigned char) *ptr))
 		ptr++;
 	*cpend = save;
 finish:
-	/* the >= should be ==, but for SunOS 3.5 strtod() */
-	if (errno == 0 && ptr >= cpend) {
-		n->flags |= newflags;
-		n->flags |= NUMCUR;
+	if (errno == 0) {
+		if (ptr == cpend) {
+			n->flags |= newflags;
+			n->flags |= NUMCUR;
+		}
+		/* else keep the leading numeric value without updating flags */
 	} else {
-		if (0 && do_lint && ptr < cpend)
-			lintwarn(_("can't convert string to float"));
 		errno = 0;
+		/*
+		 * N.B. For subnormal values, strtod may return the
+		 * floating-point representation while setting errno to ERANGE.
+		 * We force the numeric value to 0 in such cases.
+		 */
+		n->numbr = 0;
 	}
 
-	return n->numbr;
+	return n;
 }
 
 
 /*
- * the following lookup table is used as an optimization in force_string
+ * The following lookup table is used as an optimization in force_string;
  * (more complicated) variations on this theme didn't seem to pay off, but 
- * systematic testing might be in order at some point
+ * systematic testing might be in order at some point.
  */
 static const char *values[] = {
 	"0",
@@ -175,26 +177,14 @@ static const char *values[] = {
 };
 #define	NVAL	(sizeof(values)/sizeof(values[0]))
 
-/* format_val --- format a numeric value based on format */
+/* r_format_val --- format a numeric value based on format */
 
 NODE *
-format_val(const char *format, int index, register NODE *s)
+r_format_val(const char *format, int index, NODE *s)
 {
 	char buf[BUFSIZ];
-	register char *sp = buf;
+	char *sp = buf;
 	double val;
-	char *orig, *trans, save;
-
-	if (! do_traditional && (s->flags & INTLSTR) != 0) {
-		save = s->stptr[s->stlen];
-		s->stptr[s->stlen] = '\0';
-
-		orig = s->stptr;
-		trans = dgettext(TEXTDOMAIN, orig);
-
-		s->stptr[s->stlen] = save;
-		return tmp_string(trans, strlen(trans));
-	}
 
 	/*
 	 * 2/2007: Simplify our lives here. Instead of worrying about
@@ -208,14 +198,17 @@ format_val(const char *format, int index, register NODE *s)
 	 * do a lot of conversion of integers to strings. So, we reinstate
 	 * the old code, but use %.0f for integral values that are outside
 	 * the range of a long.  This seems a reasonable compromise.
+	 *
+	 * 12/2009: Use <= and >= in the comparisons with LONG_xxx instead of
+	 * < and > so that things work correctly on systems with 64 bit integers.
 	 */
 
 	/* not an integral value, or out of range */
 	if ((val = double_to_int(s->numbr)) != s->numbr
-	    || val < LONG_MIN || val > LONG_MAX) {
+			|| val <= LONG_MIN || val >= LONG_MAX
+	) {
 		/*
-		 * Once upon a time, if GFMT_WORKAROUND wasn't defined,
-		 * we just blindly did this:
+		 * Once upon a time, we just blindly did this:
 		 *	sprintf(sp, format, s->numbr);
 		 *	s->stlen = strlen(sp);
 		 *	s->stfmt = (char) index;
@@ -223,38 +216,36 @@ format_val(const char *format, int index, register NODE *s)
 		 * and just always format the value ourselves.
 		 */
 
-		NODE *dummy, *r;
-		unsigned short oflags;
-		extern NODE **fmt_list;          /* declared in eval.c */
+		NODE *dummy[2], *r;
+		unsigned int oflags;
 
 		/* create dummy node for a sole use of format_tree */
-		getnode(dummy);
-		dummy->type = Node_expression_list;
-		dummy->lnode = s;
-		dummy->rnode = NULL;
+		dummy[1] = s;
 		oflags = s->flags;
-		s->flags |= PERM; /* prevent from freeing by format_tree() */
+
 		if (val == s->numbr) {
 			/* integral value, but outside range of %ld, use %.0f */
 			r = format_tree("%.0f", 4, dummy, 2);
 			s->stfmt = -1;
 		} else {
 			r = format_tree(format, fmt_list[index]->stlen, dummy, 2);
+			assert(r != NULL);
 			s->stfmt = (char) index;
 		}
 		s->flags = oflags;
 		s->stlen = r->stlen;
 		if ((s->flags & STRCUR) != 0)
-			free(s->stptr);
+			efree(s->stptr);
 		s->stptr = r->stptr;
-		freenode(r);		/* Do not free_temp(r)!  We want */
-		freenode(dummy);	/* to keep s->stptr == r->stpr.  */
+		freenode(r);	/* Do not unref(r)! We want to keep s->stptr == r->stpr.  */
 
 		goto no_malloc;
 	} else {
-		/* integral value */
-	        /* force conversion to long only once */
-		register long num = (long) val;
+		/*
+		 * integral value; force conversion to long only once.
+		 */
+		long num = (long) val;
+
 		if (num < NVAL && num >= 0) {
 			sp = (char *) values[num];
 			s->stlen = 1;
@@ -263,76 +254,42 @@ format_val(const char *format, int index, register NODE *s)
 			s->stlen = strlen(sp);
 		}
 		s->stfmt = -1;
+		if ((s->flags & INTIND) != 0) {
+			s->flags &= ~(INTIND|NUMBER);
+			s->flags |= STRING;
+		}
 	}
+	if (s->stptr != NULL)
+		efree(s->stptr);
 	emalloc(s->stptr, char *, s->stlen + 2, "format_val");
-	memcpy(s->stptr, sp, s->stlen+1);
+	memcpy(s->stptr, sp, s->stlen + 1);
 no_malloc:
-	s->stref = 1;
 	s->flags |= STRCUR;
 	free_wstr(s);
 	return s;
 }
 
-/* r_force_string --- force a value to be a string */
-
-NODE *
-r_force_string(register NODE *s)
-{
-	NODE *ret;
-#ifdef GAWKDEBUG
-	if (s == NULL)
-		cant_happen();
-	if (s->type != Node_val)
-		cant_happen();
-	if (s->stref <= 0)
-		cant_happen();
-	if ((s->flags & STRCUR) != 0
-	    && (s->stfmt == -1 || s->stfmt == CONVFMTidx))
-		return s;
-#endif
-
-	ret = format_val(CONVFMT, CONVFMTidx, s);
-	return ret;
-}
-
-/*
- * dupnode:
- * Duplicate a node.  (For strings, "duplicate" means crank up the
- * reference count.)
- */
+/* r_dupnode --- duplicate a node */
 
 NODE *
 r_dupnode(NODE *n)
 {
-	register NODE *r;
+	NODE *r;
 
-#ifndef DUPNODE_MACRO
-	if ((n->flags & TEMP) != 0) {
-		n->flags &= ~TEMP;
-		n->flags |= MALLOC;
+	assert(n->type == Node_val);
+
+#ifdef GAWKDEBUG
+	if ((n->flags & MALLOC) != 0) {
+		n->valref++;
 		return n;
 	}
-	if ((n->flags & PERM) != 0)
-		return n;
 #endif
-	if ((n->flags & (MALLOC|STRCUR)) == (MALLOC|STRCUR)) {
-		if (n->stref < LONG_MAX)
-			n->stref++;
-		else
-			n->flags |= PERM;
-		return n;
-	} else if ((n->flags & MALLOC) != 0 && n->type == Node_ahash) {
-		if (n->ahname_ref < LONG_MAX)
-			n->ahname_ref++;
-		else
-			n->flags |= PERM;
-		return n;
-	}
+
 	getnode(r);
 	*r = *n;
-	r->flags &= ~(PERM|TEMP|FIELD);
+	r->flags &= ~FIELD;
 	r->flags |= MALLOC;
-#ifdef MBS_SUPPORT
+	r->valref = 1;
 	/*
 	 * DON'T call free_wstr(r) here!
 	 * r->wstptr still points at n->wstptr's value, and we
@@ -340,103 +297,102 @@ r_dupnode(NODE *n)
 	 */
 	r->wstptr = NULL;
 	r->wstlen = 0;
-#endif /* defined MBS_SUPPORT */
-	if (n->type == Node_val && (n->flags & STRCUR) != 0) {
-		r->stref = 1;
-		emalloc(r->stptr, char *, r->stlen + 2, "dupnode");
-		memcpy(r->stptr, n->stptr, r->stlen);
-		r->stptr[r->stlen] = '\0';
-#if defined MBS_SUPPORT
+
+	if ((n->flags & STRCUR) != 0) {
+		emalloc(r->stptr, char *, n->stlen + 2, "r_dupnode");
+		memcpy(r->stptr, n->stptr, n->stlen);
+		r->stptr[n->stlen] = '\0';
 		if ((n->flags & WSTRCUR) != 0) {
 			r->wstlen = n->wstlen;
-			emalloc(r->wstptr, wchar_t *, sizeof(wchar_t) * (r->wstlen + 2), "dupnode");
-			memcpy(r->wstptr, n->wstptr, r->wstlen * sizeof(wchar_t));
-			r->wstptr[r->wstlen] = L'\0';
+			emalloc(r->wstptr, wchar_t *, sizeof(wchar_t) * (n->wstlen + 2), "r_dupnode");
+			memcpy(r->wstptr, n->wstptr, n->wstlen * sizeof(wchar_t));
+			r->wstptr[n->wstlen] = L'\0';
 			r->flags |= WSTRCUR;
 		}
-#endif /* defined MBS_SUPPORT */
-	} else if (n->type == Node_ahash && (n->flags & MALLOC) != 0) {
-		r->ahname_ref = 1;
-		emalloc(r->ahname_str, char *, r->ahname_len + 2, "dupnode");
-		memcpy(r->ahname_str, n->ahname_str, r->ahname_len);
-		r->ahname_str[r->ahname_len] = '\0';
 	}
+	
 	return r;
 }
 
-/* copy_node --- force a brand new copy of a node to be allocated */
+/* r_make_number --- allocate a node with defined number */
 
-NODE *
-copynode(NODE *old)
+static NODE *
+r_make_number(double x)
 {
-	NODE *new;
-	int saveflags;
-
-	assert(old != NULL);
-	saveflags = old->flags;
-	old->flags &= ~(MALLOC|PERM);
-	new = dupnode(old);
-	old->flags = saveflags;
-	return new;
-}
-
-/* mk_number --- allocate a node with defined number */
-
-NODE *
-mk_number(AWKNUM x, unsigned int flags)
-{
-	register NODE *r;
-
+	NODE *r;
 	getnode(r);
 	r->type = Node_val;
 	r->numbr = x;
-	r->flags = flags;
-#ifdef GAWKDEBUG
-	r->stref = 1;
+	r->flags = MALLOC|NUMBER|NUMCUR;
+	r->valref = 1;
 	r->stptr = NULL;
 	r->stlen = 0;
-	free_wstr(r);
-#endif /* GAWKDEBUG */
+	r->wstptr = NULL;
+	r->wstlen = 0;
 	return r;
 }
 
+/* cmp_awknums --- compare two AWKNUMs */
+
+int
+cmp_awknums(const NODE *t1, const NODE *t2)
+{
+	/*
+	 * This routine is also used to sort numeric array indices or values.
+	 * For the purposes of sorting, NaN is considered greater than
+	 * any other value, and all NaN values are considered equivalent and equal.
+	 * This isn't in compliance with IEEE standard, but compliance w.r.t. NaN
+	 * comparison at the awk level is a different issue, and needs to be dealt
+	 * with in the interpreter for each opcode seperately.
+	 */
+
+	if (isnan(t1->numbr))
+		return ! isnan(t2->numbr);
+	if (isnan(t2->numbr))
+		return -1;
+	/* don't subtract, in case one or both are infinite */
+	if (t1->numbr == t2->numbr)
+		return 0;
+	if (t1->numbr < t2->numbr)
+		return -1;
+	return 1;
+}
+
+
 /* make_str_node --- make a string node */
 
 NODE *
-make_str_node(char *s, unsigned long len, int flags)
+make_str_node(const char *s, size_t len, int flags)
 {
-	register NODE *r;
-
+	NODE *r;
 	getnode(r);
 	r->type = Node_val;
-	r->flags = (STRING|STRCUR|MALLOC);
-#ifdef MBS_SUPPORT
+	r->numbr = 0;
+	r->flags = (MALLOC|STRING|STRCUR);
+	r->valref = 1;
+	r->stfmt = -1;
 	r->wstptr = NULL;
 	r->wstlen = 0;
-#endif /* defined MBS_SUPPORT */
 
-	if (flags & ALREADY_MALLOCED)
-		r->stptr = s;
+	if ((flags & ALREADY_MALLOCED) != 0)
+		r->stptr = (char *) s;
 	else {
-		emalloc(r->stptr, char *, len + 2, s);
+		emalloc(r->stptr, char *, len + 2, "make_str_node");
 		memcpy(r->stptr, s, len);
 	}
 	r->stptr[len] = '\0';
-	       
+       
 	if ((flags & SCAN) != 0) {	/* scan for escape sequences */
 		const char *pf;
-		register char *ptm;
-		register int c;
-		register const char *end;
-#ifdef MBS_SUPPORT
+		char *ptm;
+		int c;
+		const char *end;
 		mbstate_t cur_state;
 
 		memset(& cur_state, 0, sizeof(cur_state));
-#endif
 
 		end = &(r->stptr[len]);
 		for (pf = ptm = r->stptr; pf < end;) {
-#ifdef MBS_SUPPORT
 			/*
 			 * Keep multibyte characters together. This avoids
 			 * problems if a subsequent byte of a multibyte
@@ -453,7 +409,7 @@ make_str_node(char *s, unsigned long len, int flags)
 					continue;
 				}
 			}
-#endif
+
 			c = *pf++;
 			if (c == '\\') {
 				c = parse_escape(&pf);
@@ -469,107 +425,41 @@ make_str_node(char *s, unsigned long len, int flags)
 		len = ptm - r->stptr;
 		erealloc(r->stptr, char *, len + 1, "make_str_node");
 		r->stptr[len] = '\0';
-		r->flags |= PERM;
 	}
 	r->stlen = len;
-	r->stref = 1;
-	r->stfmt = -1;
 
 	return r;
 }
 
-/* tmp_string --- allocate a temporary string */
-
-NODE *
-tmp_string(char *s, size_t len)
-{
-	register NODE *r;
-
-	r = make_string(s, len);
-	r->flags |= TEMP;
-	return r;
-}
-
-/* more_nodes --- allocate more nodes */
-
-#define NODECHUNK	100
-
-NODE *nextfree = NULL;
-
-NODE *
-more_nodes()
-{
-	register NODE *np;
-
-	/* get more nodes and initialize list */
-	emalloc(nextfree, NODE *, NODECHUNK * sizeof(NODE), "more_nodes");
-	memset(nextfree, 0, NODECHUNK * sizeof(NODE));
-	for (np = nextfree; np <= &nextfree[NODECHUNK - 1]; np++) {
-		np->nextp = np + 1;
-	}
-	--np;
-	np->nextp = NULL;
-	np = nextfree;
-	nextfree = nextfree->nextp;
-	return np;
-}
-
-#ifdef MEMDEBUG
-#undef freenode
-/* freenode --- release a node back to the pool */
-
-void
-freenode(NODE *it)
-{
-#ifdef MPROF
-	it->stref = 0;
-	free((char *) it);
-#else	/* not MPROF */
-#ifndef NO_PROFILING
-	it->exec_count = 0;
-#endif
-	/* add it to head of freelist */
-	it->nextp = nextfree;
-	nextfree = it;
-#endif	/* not MPROF */
-}
-#endif	/* GAWKDEBUG */
 
 /* unref --- remove reference to a particular node */
 
 void
-unref(register NODE *tmp)
+r_unref(NODE *tmp)
 {
+#ifdef GAWKDEBUG
 	if (tmp == NULL)
 		return;
-	if ((tmp->flags & PERM) != 0)
-		return;
-	tmp->flags &= ~TEMP;
 	if ((tmp->flags & MALLOC) != 0) {
-		if (tmp->type == Node_ahash) {
-			if (tmp->ahname_ref > 1) {
-				tmp->ahname_ref--;
-				return;
-			}
-			free(tmp->ahname_str);
-		} else if ((tmp->flags & STRCUR) != 0) {
-			if (tmp->stref > 1) {
-				tmp->stref--;
-				return;
-			}
-			free(tmp->stptr);
-			free_wstr(tmp);
+		if (tmp->valref > 1) {
+			tmp->valref--;
+			return;
 		}
-		freenode(tmp);
-		return;
-	}
-	if ((tmp->flags & FIELD) != 0) {
-		free_wstr(tmp);
-		freenode(tmp);
-		return;
+		if ((tmp->flags & STRCUR) != 0)
+			efree(tmp->stptr);
 	}
+#else
+	if ((tmp->flags & (MALLOC|STRCUR)) == (MALLOC|STRCUR))
+		efree(tmp->stptr);
+#endif
+
+	mpfr_unset(tmp);
+
+	free_wstr(tmp);
+	freenode(tmp);
 }
 
+
 /*
  * parse_escape:
  *
@@ -586,20 +476,21 @@ unref(register NODE *tmp)
  * If \ is followed by 000, we return 0 and leave the string pointer after the
  * zeros.  A value of 0 does not mean end of string.  
  *
- * Posix doesn't allow \x.
+ * POSIX doesn't allow \x.
  */
 
 int
 parse_escape(const char **string_ptr)
 {
-	register int c = *(*string_ptr)++;
-	register int i;
-	register int count;
+	int c = *(*string_ptr)++;
+	int i;
+	int count;
 	int j;
 	const char *start;
 
 	if (do_lint_old) {
 		switch (c) {
+		case 'a':
 		case 'b':
 		case 'f':
 		case 'r':
@@ -610,7 +501,7 @@ parse_escape(const char **string_ptr)
 
 	switch (c) {
 	case 'a':
-		return BELL;
+		return '\a';
 	case 'b':
 		return '\b';
 	case 'f':
@@ -650,16 +541,16 @@ parse_escape(const char **string_ptr)
 		return i;
 	case 'x':
 		if (do_lint) {
-			static short warned = FALSE;
+			static bool warned = false;
 
 			if (! warned) {
-				warned = TRUE;
+				warned = true;
 				lintwarn(_("POSIX does not allow `\\x' escapes"));
 			}
 		}
 		if (do_posix)
 			return ('x');
-		if (! ISXDIGIT((*string_ptr)[0])) {
+		if (! isxdigit((unsigned char) (*string_ptr)[0])) {
 			warning(_("no hex digits in `\\x' escape sequence"));
 			return ('x');
 		}
@@ -668,11 +559,11 @@ parse_escape(const char **string_ptr)
 		for (;; j++) {
 			/* do outside test to avoid multiple side effects */
 			c = *(*string_ptr)++;
-			if (ISXDIGIT(c)) {
+			if (isxdigit(c)) {
 				i *= 16;
-				if (ISDIGIT(c))
+				if (isdigit(c))
 					i += c - '0';
-				else if (ISUPPER(c))
+				else if (isupper(c))
 					i += c - 'A' + 10;
 				else
 					i += c - 'a' + 10;
@@ -689,13 +580,13 @@ parse_escape(const char **string_ptr)
 		return c;
 	default:
 	{
-		static short warned[256];
+		static bool warned[256];
 		unsigned char uc = (unsigned char) c;
 
 		/* N.B.: use unsigned char here to avoid Latin-1 problems */
 
 		if (! warned[uc]) {
-			warned[uc] = TRUE;
+			warned[uc] = true;
 
 			warning(_("escape sequence `\\%c' treated as plain `%c'"), uc, uc);
 		}
@@ -704,12 +595,14 @@ parse_escape(const char **string_ptr)
 	}
 }
 
-/* isnondecimal --- return true if number is not a decimal number */
+/* get_numbase --- return the base to use for the number in 's' */
 
 int
-isnondecimal(const char *str, int use_locale)
+get_numbase(const char *s, bool use_locale)
 {
 	int dec_point = '.';
+	const char *str = s;
+
 #if defined(HAVE_LOCALE_H)
 	/*
 	 * loc.decimal_point may not have been initialized yet,
@@ -720,11 +613,11 @@ isnondecimal(const char *str, int use_locale)
 #endif
 
 	if (str[0] != '0')
-		return FALSE;
+		return 10;
 
 	/* leading 0x or 0X */
 	if (str[1] == 'x' || str[1] == 'X')
-		return TRUE;
+		return 16;
 
 	/*
 	 * Numbers with '.', 'e', or 'E' are decimal.
@@ -734,15 +627,18 @@ isnondecimal(const char *str, int use_locale)
 	 */
 	for (; *str != '\0'; str++) {
 		if (*str == 'e' || *str == 'E' || *str == dec_point)
-			return FALSE;
-		else if (! ISDIGIT(*str))
+			return 10;
+		else if (! isdigit((unsigned char) *str))
 			break;
 	}
 
-	return TRUE;
+	if (! isdigit((unsigned char) s[1])
+			|| s[1] == '8' || s[1] == '9'
+	)
+		return 10;
+	return 8;
 }
 
-#if defined MBS_SUPPORT
 /* str2wstr --- convert a multibyte string to a wide string */
 
 NODE *
@@ -752,9 +648,19 @@ str2wstr(NODE *n, size_t **ptr)
 	char *sp;
 	mbstate_t mbs;
 	wchar_t wc, *wsp;
+	static bool warned = false;
 
 	assert((n->flags & (STRING|STRCUR)) != 0);
 
+	/*
+	 * Don't convert global null string or global null field
+	 * variables to a wide string. They are both zero-length anyway.
+	 * This also avoids future double-free errors while releasing
+	 * shallow copies, eg. *tmp = *Null_field; free_wstr(tmp);
+	 */
+	if (n == Nnull_string || n == Null_field)
+		return n;
+
 	if ((n->flags & WSTRCUR) != 0) {
 		if (ptr == NULL)
 			return n;
@@ -768,11 +674,11 @@ str2wstr(NODE *n, size_t **ptr)
 	 * code trades space for time. We allocate
 	 * an array of wchar_t that is n->stlen long.
 	 * This is needed in the worst case anyway, where
-	 * each input bytes maps to one wchar_t.  The
+	 * each input byte maps to one wchar_t.  The
 	 * advantage is that we only have to convert the string
 	 * once, instead of twice, once to find out how many
-	 * wide characters, and then again to actually fill
-	 * the info in.  If there's a lot left over, we can
+	 * wide characters, and then again to actually fill in
+	 * the info.  If there's a lot left over, we can
 	 * realloc the wide string down in size.
 	 */
 
@@ -796,11 +702,39 @@ str2wstr(NODE *n, size_t **ptr)
 	src_count = n->stlen;
 	memset(& mbs, 0, sizeof(mbs));
 	for (i = 0; src_count > 0; i++) {
-		count = mbrtowc(& wc, sp, src_count, & mbs);
+		/*
+		 * 9/2010: Check the current byte; if it's a valid character,
+		 * then it doesn't start a multibyte sequence. This brings a
+		 * big speed up. Thanks to Ulrich Drepper for the tip.
+		 * 11/2010: Thanks to Paolo Bonzini for some even faster code.
+		 */
+		if (is_valid_character(*sp)) {
+			count = 1;
+			wc = btowc_cache(*sp);
+		} else
+			count = mbrtowc(& wc, sp, src_count, & mbs);
 		switch (count) {
 		case (size_t) -2:
 		case (size_t) -1:
-			goto done;
+			/*
+			 * Just skip the bad byte and keep going, so that
+			 * we get a more-or-less full string, instead of
+			 * stopping early. This is particularly important
+			 * for match() where we need to build the indices.
+			 */
+			sp++;
+			src_count--;
+			/*
+			 * mbrtowc(3) says the state of mbs becomes undefined
+			 * after a bad character, so reset it.
+			 */
+			memset(& mbs, 0, sizeof(mbs));
+			/* And warn the user something's wrong */
+			if (do_lint && ! warned) {
+				warned = true;
+				lintwarn(_("Invalid multibyte data detected. There may be a mismatch between your data and your locale."));
+			}
+			break;
 
 		case 0:
 			count = 1;
@@ -817,9 +751,8 @@ str2wstr(NODE *n, size_t **ptr)
 		}
 	}
 
-done:
 	*wsp = L'\0';
-	n->wstlen = i;
+	n->wstlen = wsp - n->wstptr;
 	n->flags |= WSTRCUR;
 #define ARBITRARY_AMOUNT_TO_GIVE_BACK 100
 	if (n->stlen - n->wstlen > ARBITRARY_AMOUNT_TO_GIVE_BACK)
@@ -828,23 +761,64 @@ done:
 	return n;
 }
 
+/* wstr2str --- convert a wide string back into multibyte one */
+
+NODE *
+wstr2str(NODE *n)
+{
+	size_t result;
+	size_t length;
+	wchar_t *wp;
+	mbstate_t mbs;
+	char *newval, *cp;
+
+	assert(n->valref == 1);
+	assert((n->flags & WSTRCUR) != 0);
+
+	/*
+	 * Convert the wide chars in t1->wstptr back into m.b. chars.
+	 * This is pretty grotty, but it's the most straightforward
+	 * way to do things.
+	 */
+	memset(& mbs, 0, sizeof(mbs));
+
+	length = n->wstlen;
+	emalloc(newval, char *, (length * gawk_mb_cur_max) + 2, "wstr2str");
+
+	wp = n->wstptr;
+	for (cp = newval; length > 0; length--) {
+		result = wcrtomb(cp, *wp, & mbs);
+		if (result == (size_t) -1)	/* what to do? break seems best */
+			break;
+		cp += result;
+		wp++;
+	}
+	*cp = '\0';
+
+	efree(n->stptr);
+	n->stptr = newval;
+	n->stlen = cp - newval;
+
+	return n;
+}
+
 /* free_wstr --- release the wide string part of a node */
 
 void
-free_wstr(NODE *n)
+r_free_wstr(NODE *n)
 {
 	assert(n->type == Node_val);
 
 	if ((n->flags & WSTRCUR) != 0) {
 		assert(n->wstptr != NULL);
-		free(n->wstptr);
+		efree(n->wstptr);
 	}
 	n->wstptr = NULL;
 	n->wstlen = 0;
 	n->flags &= ~WSTRCUR;
 }
 
-static void
+static void __attribute__ ((unused))
 dump_wstr(FILE *fp, const wchar_t *str, size_t len)
 {
 	if (str == NULL || len == 0)
@@ -913,7 +887,6 @@ out:	;
 
 	return NULL;
 }
-#endif /* defined MBS_SUPPORT */
 
 /* is_ieee_magic_val --- return true for +inf, -inf, +nan, -nan */
 
@@ -938,7 +911,7 @@ is_ieee_magic_val(const char *val)
 static AWKNUM
 get_ieee_magic_val(const char *val)
 {
-	static short first = TRUE;
+	static bool first = true;
 	static AWKNUM inf;
 	static AWKNUM nan;
 
@@ -947,7 +920,7 @@ get_ieee_magic_val(const char *val)
 
 	if (val == ptr) { /* Older strtod implementations don't support inf or nan. */
 		if (first) {
-			first = FALSE;
+			first = false;
 			nan = sqrt(-1.0);
 			inf = -log(0.0);
 		}
@@ -959,3 +932,54 @@ get_ieee_magic_val(const char *val)
 
 	return v;
 }
+
+wint_t btowc_cache[256];
+
+/* init_btowc_cache --- initialize the cache */
+
+void init_btowc_cache()
+{
+	int i;
+
+	for (i = 0; i < 255; i++) {
+		btowc_cache[i] = btowc(i);
+	}
+}
+
+#define BLOCKCHUNK 100
+
+BLOCK nextfree[BLOCK_MAX] = {
+	{ 0, NULL},	/* invalid */	
+	{ sizeof(NODE), NULL },
+	{ sizeof(BUCKET), NULL },
+};
+
+
+/* more_blocks --- get more blocks of memory and add to the free list;
+	size of a block must be >= sizeof(BLOCK)
+ */
+
+void *
+more_blocks(int id)
+{
+	BLOCK *freep, *np, *next;
+	char *p, *endp;
+	size_t size;
+
+	size = nextfree[id].size;
+
+	emalloc(freep, BLOCK *, BLOCKCHUNK * size, "more_blocks");
+	p = (char *) freep;
+	endp = p + BLOCKCHUNK * size;
+
+	for (np = freep; ; np = next) {
+		next = (BLOCK *) (p += size);
+		if (p >= endp) {
+			np->freep = NULL;
+			break;
+		}
+		np->freep = next;
+	}
+	nextfree[id].freep = freep->freep;
+	return freep;
+}
